### Added feature: HTTPS to MLLP
### By Tiago Rodrigues
### Sectra Iberia, Aug 2022
### Addapted from https://github.com/rivethealth/mllp-http
import base64
import functools
import http.server
import logging
import os
import socket
import ssl
import threading
import time
from datetime import datetime, timezone

from .mllp import send_mllp, parse_mllp

logger = logging.getLogger(__name__)

class MllpClientOptions:
    def __init__(self, keep_alive, max_messages, timeout):
        # self.address = address
        self.keep_alive = keep_alive
        self.max_messages = max_messages
        self.timeout = timeout


class MllpClient:
    def __init__(self, address, options):
        self.address = address
        self.options = options
        self.connections = []
        self.lock = threading.Lock()

    def _check_connection(self, connection):
        while not connection.closed:
            elasped = (
                connection.last_update - time.monotonic()
                if connection.last_update is not None
                else 0
            )
            remaining = self.options.keep_alive + elasped
            if 0 < remaining:
                time.sleep(remaining)
            else:
                try:
                    with self.lock:
                        self.connections.remove(connection)
                except ValueError:
                    pass
                else:
                    if self.options.keep_alive > 0:
                        # To keep the connection alive in case keep_alive is -1
                        connection.close()

    def _connect(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.options.timeout:
            s.settimeout(self.options.timeout)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 10)
        # print(self.address)
        s.connect(self.address)
        connection = MllpConnection(s)
        if self.options.keep_alive is not None:
            thread = threading.Thread(
                daemon=False,
                target=self._check_connection,
                args=(connection,)
            )
            thread.start()
        return connection

    def send(self, data):
        with self.lock:
            try:
                connection = self.connections.pop()
            except IndexError:
                connection = None
            else:
                connection.last_update = None
        if connection is None:
            connection = self._connect()
        response = connection.send(data)
        if self.options.max_messages <= connection.message_count and self.options.max_messages >= 0:
            connection.close()
        else:
            connection.last_update = time.monotonic()
            with self.lock:
                self.connections.append(connection)
        return response


class MllpConnection:
    def __init__(self, socket):
        self.closed = False
        self.last_update = None
        self.message_count = 0
        self.socket = socket

    def close(self):
        self.close = True
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        logger.info("Disconnected from MLLP Server")
        #print("Disconnected from MLLP Server")

    def send(self, data):
        self.message_count += 1

        # To send the HL7 messages, it will make use of an MLLP parser to format the data
        # The parser will return the ACK/NACK response
        return send_mllp(self.socket, data)


class HttpsServerOptions:
    def __init__(self, timeout, content_type, certfile, keyfile, keep_alive, username, password, mllp_parser):
        self.timeout = timeout
        self.content_type = content_type
        self.certfile = certfile
        self.keyfile = keyfile
        self.keep_alive = keep_alive
        self.username = username
        self.password = password
        self.mllp_parser = mllp_parser


class Authentication:
    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.key = ''
        self.key = base64.b64encode(
            bytes('%s:%s' % (self.username, self.password), 'utf-8')).decode('ascii')

    def get_auth_key(self):
        return self.key


class HttpsHandler(http.server.BaseHTTPRequestHandler):
    def __init__(self, request, address, server, mllp_client,
                 content_type, timeout, keep_alive, authentication, mllp_parser):
        self.protocol_version = "HTTP/1.1"
        self.content_type = content_type
        self.mllp_client = mllp_client
        self.timeout = timeout
        self.keep_alive = keep_alive
        self.authentication = authentication
        self.mllp_parser = mllp_parser
        super().__init__(request, address, server)

    # For Authentication
    def do_AUTHHEAD(self):
        self.send_response(401)
        self.send_header('WWW-Authenticate', b'Basic realm=\"Test\"')
        self.send_header('Content-type', b'text/html')
        self.end_headers()

    # To check auth on server and client side
    def check_auth(self):
        key = self.authentication.get_auth_key()
        # print(self.headers['Authorization'])
        # print(self.headers['Content-type'])
        # print(self.headers['Date'])
        if self.headers['Authorization'] == None:
            self.do_AUTHHEAD()
            self.wfile.write(b'Authentication is required.')
            logger.info('Client did not authenticate.')
            return False
        elif self.headers['Authorization'] == 'Basic ' + str(key):
            logger.info('Client Authenticated.')
            return True
        else:
            self.do_AUTHHEAD()
            self.wfile.write(bytes(self.headers['Authorization'], "ascii"))
            self.wfile.write(b'Authentication failed: Wrong credentials.')
            logger.info('Client did not provide the correct credentials.')
            return False

    def do_POST(self):
        try:
            # Check authentication
            authentication_step = False
            if self.authentication != None:
                try:
                    # if username and password were given as argument, client should authenticate to server
                    authentication_step = self.check_auth()
                except Exception as e:
                    logger.error("Authentication error: %s", e)
            else:
                # if username and password were not given as argument, do not need to authenticate to server
                authentication_step = True

            if authentication_step:
                # Process received data
                content_length = int(self.headers["Content-Length"])  # From the received data
                data = self.rfile.read(content_length)  # Read income data
                logger.info("Message: %s bytes", len(data))
                logger.info("Received Data:\n{}".format(data))

                # Send data to MLLP Listener and get MLLP ACK response:
                response = self.mllp_client.send(data)

                # Parse the data from the MLLP response:
                #   > Remove Start block and End Block
                #   > Correct Carriage Return if needed
                if self.mllp_parser:
                    response = parse_mllp(response)

                # Prepare and send back ACK to HTTPS client
                self.send_response(200)
                self.send_header("Content-Length", len(response))
                if self.content_type:
                    self.send_header("Content-Type", self.content_type)
                    #print(self.content_type)
                if self.keep_alive is not None:
                    self.send_header("Keep-Alive", f"timeout={self.keep_alive}")
                now = datetime.now(timezone.utc)
                date = now.strftime("%a, %d %b %y %H:%M:%S %Z")
                # print(date)
                self.send_header("Date", date)
                self.end_headers()
                self.wfile.write(response)
                logger.info("Response: %s bytes", len(response))
                logger.info("Response Data: 200 Ok\n{}".format(response))

        except Exception as e:
            logger.error("HTTPS connection error: %s", e)


def serve(address, options, mllp_address, mllp_options):

    # MLLP Client for dealing with the MLLP TCP connection
    client = MllpClient(mllp_address, mllp_options)

    # If authentication is enabled, set the auth
    if options.username and options.password:
        auth = Authentication(
            username=options.username,
            password=options.password,
        )
        logger.info(
            "Authentication on server side is enabled.\nUser: {}\nPassword: {}".format(options.username, "*"*len(options.password)))
    elif os.environ.get("HTTP_AUTHORIZATION"):
        auth = os.environ["HTTP_AUTHORIZATION"]
        logger.info(
            "Authentication on server side is enabled with Environment Variable HTTP_AUTHORIZATION.")
    else:
        auth = None
        logger.warning(
            "Authentication on server side is disabled.")

    # HTTP server handler
    handler = functools.partial(
        HttpsHandler,
        content_type=options.content_type,
        keep_alive=options.keep_alive,
        timeout=options.timeout or None,
        mllp_client=client,
        authentication=auth,
        mllp_parser=options.mllp_parser,
    )


    try:
        server = http.server.ThreadingHTTPServer(address, handler)

        # >> Dealing with SSL/TLS on the HTTP server side
        # For Python > 3.7
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.load_cert_chain(
            certfile=options.certfile,
            keyfile=options.keyfile,
        )
        server.socket = context.wrap_socket(
            server.socket,
            server_side=True,
        )
        # For Python < 3.2 SSLContext is not used, change it by ssl.wrap_socket:
        # server.socket = ssl.wrap_socket(
        #     server.socket,
        #     server_side=True,
        #     certfile="C:/ssl/certfile.crt",
        #     keyfile="C:/ssl/keyfile.key",
        # )

        logger.info("Listening on %s:%s", address[0], address[1])
        logger.info("Sending to %s:%s", mllp_address[0], mllp_address[1])
        print("\nListening on {}:{}".format(address[0], address[1]))
        print("Sending to {}:{}".format(mllp_address[0], mllp_address[1]))
        server.protocol_version = "HTTP/1.1"
        server.serve_forever()

    except Exception as e:
        logger.error("HTTPS connection error: %s", e)



### Added feature: MLLP to HTTPS
### By Tiago Rodrigues
### Sectra Iberia, Aug 2022
### Addapted from https://github.com/rivethealth/mllp-http
import base64
import datetime
import functools
import logging
import os
from datetime import datetime, timezone

import requests
import socket
import socketserver
import urllib

from requests.auth import HTTPBasicAuth

from .mllp import read_mllp, write_mllp
from .net import read_socket_bytes
from .version import __version__

logger = logging.getLogger(__name__)


def display_address(address):
    return f"{address[0]}:{address[1]}"


class MllpServerOptions:
    def __init__(self, timeout):
        self.timeout = timeout


class MllpHandler(socketserver.StreamRequestHandler):
    # Class for parsing HL7 data as MLLP Server and send it to a host as a HTTPS Client

    def __init__(self, request, address, server, timeout, https_url, https_options):
        self.https_url = https_url
        self.https_options = https_options
        self.timeout = timeout

        # If username and password are provided as arguments, use authentication
        self.username = https_options.username
        self.password = https_options.password
        self.auth = None
        if self.username and self.password:
            self.auth = ''
            self.auth = 'Basic ' + base64.b64encode(
                bytes('%s:%s' % (self.username, self.password), 'utf-8')).decode('ascii')
        super().__init__(request, address, server)

    def handle(self):
        if self.timeout:
            self.request.settimeout(self.timeout)
        self.request.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 10)

        # Session for HTTPS persistent connection
        session = requests.Session()

        # >> Notes on dealing with SSL:
        # Requests verifies SSL certificates for HTTPS requests.
        # By default, SSL verification is enabled, and Requests will
        # throw a SSLError if it’s unable to verify the certificate
        #
        # The private key to the local certificate must be unencrypted.
        # Currently, Requests does not support using encrypted keys.
        #
        # https://requests.readthedocs.io/en/latest/user/advanced/

        local_address = self.request.getsockname()
        remote_address = self.request.getpeername()

        stream = read_socket_bytes(self.rfile)

        try:
            for message in read_mllp(stream):
                try:
                    logger.info("Message: %s bytes", len(message))
                    logger.info("Received Data:\n{}\n\n".format(message))
                    headers = {
                        "Forwarded": f"by={display_address(local_address)};for={display_address(remote_address)};proto=mllp",
                        "User-Agent": f"mllp2http/{__version__}",
                        "X-Forwarded-For": display_address(remote_address),
                        "X-Forwarded-Proto": "mllp",
                    }

                    if self.auth:
                        headers["Authorization"] = self.auth
                        #print("Authorization: {}".format(self.auth))
                    elif os.environ.get("HTTP_AUTHORIZATION"):
                        headers["Authorization"] = os.environ["HTTP_AUTHORIZATION"]
                        #print("Authorization: {}".format(os.environ["HTTP_AUTHORIZATION"]))
                    if os.environ.get("API_KEY"):
                        headers["X-API-KEY"] = os.environ["API_KEY"]
                        # print(os.environ["API_KEY"])

                    if self.https_options.content_type is not None:
                        headers["Content-Type"] = self.https_options.content_type

                    now = datetime.now(timezone.utc)
                    date = now.strftime("%a, %d %b %y %H:%M:%S %Z")
                    headers["Date"] = date

                    # Disabled the warning because it was causing the program to freeze
                    if not self.https_options.verify:
                        logger.warning("Verify SSL: " + str(self.https_options.verify))
                        requests.packages.urllib3.disable_warnings()

                    # Sending the HL7 data by HTTPS by POST Method
                    response = session.post(
                        urllib.parse.urlunparse(self.https_url),
                        data=message,
                        headers=headers,
                        timeout=self.https_options.timeout,
                        verify=self.https_options.verify,  # To verify server SSL/TLS certificate. Keep as False only
                        # when using a self-sign certificate since the library does not allow self-signed certificates
                    )
                    response.raise_for_status()  # Get ACK response
                except requests.exceptions.HTTPError as e:
                    logger.error("HTTPS response error: %s", e.response.status_code)
                    break
                except Exception as e:
                    logger.error("HTTPS connection error: %s", e)
                    break
                else:
                    content = response.content
                    # print(content.decode())
                    logger.info("Response: %s bytes", len(content))
                    logger.info("Response Data:\n{}\n\n".format(content.decode()))
                    write_mllp(self.wfile, content)
                    self.wfile.flush()
        except ConnectionResetError as e:
            # An enception was raised when the MLLP client closed the connection on their side
            logger.info("MLLP Server Disconnected")
        except Exception as e:
            logger.error("Failed read MLLP message: %s", e)


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    allow_reuse_address = True


class HttpsClientOptions:
    def __init__(self, content_type, timeout, verify, username, password):
        self.content_type = content_type
        self.timeout = timeout
        self.verify = True
        if verify == "False":
            self.verify = False
        elif verify == "True":
            self.verify = True
        else:
            self.verify = verify
        self.username = username
        self.password = password


def serve(address, options, https_url, https_options):
    logger = logging.getLogger(__name__)

    # Handler for parsing the data from the MLLP socket
    handler = functools.partial(
        MllpHandler,
        https_url=https_url,
        https_options=https_options,
        timeout=options.timeout or None,
    )

    try:
        # MLLP Server/Listener
        server = ThreadedTCPServer(address, handler)

        logger.info("Listening on %s:%s", address[0], address[1])
        logger.info("Sending to %s", https_url[1])
        print("\nListening on {}:{}".format(address[0], address[1]))
        print("Sending to {}".format(https_url[1]))
        server.serve_forever()

    except Exception as e:
        logger.error("MLLP connection error: %s", e)



### By Tiago Rodrigues
### Sectra Iberia, Aug 2022
### Script to manage the logging


import logging
import logging.handlers as handlers
from time import sleep
from threading import Thread
import os
import sys
import datetime


# Generates and manages the logs files
class Log2File:
    def __init__(self, file_name, folder_path, log_level_str, number_of_days_log=1):
        self.file_name = file_name
        self.folder_path = folder_path
        self.log_level_str = log_level_str
        self.number_of_days_log = number_of_days_log

    def log_level(self, arg):
        if arg == "error":
            return logging.ERROR
        elif arg == "warn":
            return logging.WARNING
        elif arg == "info":
            return logging.INFO

    def new_log(self):
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(name)s %(message)s"
        )
        level = self.log_level(self.log_level_str)

        logger = logging.getLogger()
        logger.setLevel(level)

        filename = self.folder_path + "\\" + self.file_name
        log_handler = handlers.TimedRotatingFileHandler(filename, when='D', interval=self.number_of_days_log)
        log_handler.setFormatter(formatter)

        logger.addHandler(log_handler)


# Looks and deletes old
class LogMonitor:
    def __init__(self, number_of_days_check, folder_path):
        self.daemon = Thread(target=self.background_task, args=(number_of_days_check,), daemon=True, name='LogControl')
        self.folder_path = folder_path

    # task that runs at a fixed interval
    def background_task(self, interval_days):
        interval_sec = interval_days * 24 * 60 * 60  # Days to sec

        # run forever
        while True:
            # block for the interval
            sleep(interval_sec)

            # perform the task
            N = interval_days
            if not os.path.exists(self.folder_path):
                print("Please provide valid path")
                sys.exit(1)
            if os.path.isfile(self.folder_path):
                print("Please provide dictionary path")
                sys.exit(2)
            today = datetime.datetime.now()
            for each_file in os.listdir(self.folder_path):
                if each_file[len(each_file) - 4:len(each_file)] != ".log":
                    each_file_path = os.path.join(self.folder_path, each_file)
                    if os.path.isfile(each_file_path):
                        datetime_str = each_file[len(each_file) - 19:len(each_file)]

                        file_cre_date = datetime.datetime.strptime(datetime_str, '%Y-%m-%d_%H-%M-%S')
                        # file_cre_date = datetime.datetime.fromtimestamp(os.path.getctime(each_file_path))

                        dif_days = (today - file_cre_date).days
                        if dif_days > N:
                            os.remove(each_file_path)
                            print(each_file_path, dif_days)

    def start(self):
        self.daemon.start()
        pass
